from .types import ChatPayload, FunctionResponse, StreamingChunk
from typing import Generator


class Functions:
    def __init__(self, http_client):
        self.http_client = http_client

    def chat(
        self, function_path, data: ChatPayload, stream=False
    ) -> [FunctionResponse, Generator[StreamingChunk, None, None]]:
        if stream:
            return self._chat_stream(function_path, data)
        serialized_data = data.model_dump()

        data = self.http_client.do_request(
            "POST",
            f"/v1/chat/{function_path}",
            json=serialized_data,
        )
        return FunctionResponse(**data)

    def _chat_stream(
        self,
        function_path,
        data: ChatPayload,
    ) -> Generator[StreamingChunk, None, None]:
        serialized_data = data.model_dump()
        gen = self.http_client.stream(
            "POST",
            f"/v1/chat/{function_path}?stream=True",
            json=serialized_data,
            params={"stream": "True"},
        )
        for item in gen:
            yield StreamingChunk(**item)


class AsyncFunctions:
    def __init__(self, http_client):
        self.http_client = http_client

    async def chat(
        self, function_path, data: ChatPayload, stream=False
    ) -> FunctionResponse:
        if stream:
            return self._chat_stream(function_path, data)
        serialized_data = data.model_dump()

        data = await self.http_client.do_request(
            "POST",
            f"/v1/chat/{function_path}",
            json=serialized_data,
            params={"stream": "true"} if stream else None,
        )
        return FunctionResponse(**data)

    async def _chat_stream(
        self,
        function_path,
        data: ChatPayload,
    ) -> FunctionResponse:
        gen = self.http_client.stream("POST", f"/v1/chat/{function_path}?stream=True")
        async for item in gen:
            yield StreamingChunk(**item)



import pytest
from opperai import AsyncClient, Client
from opperai.types import ChatPayload, Message

from unittest.mock import patch

api_key = "key"


@patch("opperai._client._http_client.do_request")
def test_chat(mock_do_request):
    mock_do_request.return_value = {"message": "Bonjour"}
    client = Client(api_key=api_key)
    resp = client.functions.chat(
        "french", ChatPayload(messages=[Message(role="user", content="hello")])
    )

    assert resp.message == "Bonjour"
    mock_do_request.assert_called_once()


@pytest.mark.asyncio
@patch("opperai._client._async_http_client.do_request")
async def test_async_chat(mock_do_request):
    mock_do_request.return_value = {"message": "Bonjour"}
    client = AsyncClient(api_key="op-dev-api-key", api_url="http://localhost:8000")
    resp = await client.functions.chat(
        "french", ChatPayload(messages=[Message(role="user", content="hello")])
    )

    assert resp.message == "Bonjour"
    mock_do_request.assert_called_once()


@pytest.mark.asyncio
@patch("opperai._client._async_http_client.stream")
async def test_async_chat_stream(mock_stream):
    async def gen():
        yield {"delta": "Bon"}
        yield {"delta": "Jour"}

    mock_stream.return_value = gen()
    client = AsyncClient(api_key="op-dev-api-key", api_url="http://localhost:8000")
    gen = await client.functions.chat(
        "french",
        ChatPayload(messages=[Message(role="user", content="hello")]),
        stream=True,
    )
    resp = ""
    async for message in gen:
        resp += message.delta

    assert resp == "BonJour"


@pytest.mark.asyncio
@patch("opperai._client._http_client.stream")
async def test_chat_stream(mock_stream):
    def gen():
        yield {"delta": "Bon"}
        yield {"delta": "Jour"}

    mock_stream.return_value = gen()
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    gen = client.functions.chat(
        "french",
        ChatPayload(messages=[Message(role="user", content="hello")]),
        stream=True,
    )
    resp = ""
    for message in gen:
        resp += message.delta

    assert resp == "BonJour"



from pydantic import BaseModel, Field
from typing import List, Optional, Union


class Message(BaseModel):
    role: str
    content: str


class ChatPayload(BaseModel):
    messages: List[Message]

    class Config:
        extra = "allow"


class FileMetadata(BaseModel):
    file_name: str


class ContextData(BaseModel):
    dataset_id: int
    content: str
    metadata: Union[Optional[dict], FileMetadata] = Field(default=None)


class StreamingChunk(BaseModel):
    delta: Optional[str] = None
    error: Optional[str] = None
    context: Optional[List[ContextData]] = None


class FunctionResponse(BaseModel):
    message: str
    error: Optional[str] = None
    context: Optional[List[ContextData]] = None



import httpx
from ._functions import Functions, AsyncFunctions
from httpx_sse import aconnect_sse, connect_sse
import json

DEFAULT_API_URL = "https://api.opper.ai"

Timeout = httpx.Timeout(60.0)


class AsyncClient:
    def __init__(self, api_key: str, api_url: str = DEFAULT_API_URL):
        self.http_client = _async_http_client(api_key, api_url)
        self.functions = AsyncFunctions(self.http_client)


class Client:
    def __init__(self, api_key: str, api_url: str = DEFAULT_API_URL):
        self.http_client = _http_client(api_key, api_url)
        self.functions = Functions(self.http_client)


class _async_http_client:
    def __init__(self, api_key: str, api_url):
        self.session = httpx.AsyncClient(
            base_url=api_url,
            headers={"X-OPPER-API-KEY": f"{api_key}"},
            timeout=Timeout,
        )

    async def do_request(self, method: str, path: str, **kwargs):
        response = await self.session.request(method, path, **kwargs)
        if response.status_code != 200:
            raise Exception(f"Request failed with status {response.status_code}")
        return response.json()

    async def stream(self, method: str, path: str):
        with aconnect_sse(
            self.session,
            method,
            path,
        ) as event_source:
            for sse in event_source.iter_sse():
                yield json.loads(sse.data)


class _http_client:
    def __init__(self, api_key: str, api_url: str):
        self.session = httpx.Client(
            base_url=api_url,
            headers={"X-OPPER-API-KEY": f"{api_key}"},
            timeout=Timeout,
        )

    def do_request(self, method: str, path: str, **kwargs):
        response = self.session.request(method, path, **kwargs)
        if response.status_code != 200:
            raise Exception(f"Request failed with status {response.status_code}")
        return response.json()

    def stream(self, method: str, path: str, **kwargs):
        with connect_sse(
            self.session,
            method,
            path,
            **kwargs,
        ) as event_source:
            for sse in event_source.iter_sse():
                yield json.loads(sse.data)
